"""HasAttrs: Check if objects have the same attributes as collections.abc types.

This package provides functions to check if objects have the required attributes
for various abstract base classes from collections.abc and typing, without using
isinstance checks. This is useful for duck-typing validation.

Example:
    >>> from hasattrs import has_mapping_attrs
    >>> class MyDict:
    ...     def __getitem__(self, key): pass
    ...     def __iter__(self): pass
    ...     def __len__(self): pass
    ...     # ... other Mapping methods
    >>> has_mapping_attrs(MyDict())
    True
"""

from hasattrs.checks import (
    has_abc_attrs,
    has_async_generator_attrs,
    has_async_iterable_attrs,
    has_async_iterator_attrs,
    has_attrs,
    has_awaitable_attrs,
    has_byte_string_attrs,
    has_callable_attrs,
    has_collection_attrs,
    has_container_attrs,
    has_coroutine_attrs,
    has_generator_attrs,
    has_hashable_attrs,
    has_item_view_attrs,
    has_iterable_attrs,
    has_iterator_attrs,
    has_keys_view_attrs,
    has_mapping_attrs,
    has_mapping_view_attrs,
    has_mutable_mapping_attrs,
    has_mutable_sequence_attrs,
    has_mutable_set_attrs,
    has_reversible_attrs,
    has_sequence_attrs,
    has_set_attrs,
    has_sized_attrs,
    has_values_view_attrs,
)

__version__ = "0.1.0"

__all__ = [
    "__version__",
    "has_attrs",
    "has_abc_attrs",
    "has_container_attrs",
    "has_hashable_attrs",
    "has_iterable_attrs",
    "has_iterator_attrs",
    "has_reversible_attrs",
    "has_generator_attrs",
    "has_sized_attrs",
    "has_callable_attrs",
    "has_collection_attrs",
    "has_sequence_attrs",
    "has_mutable_sequence_attrs",
    "has_byte_string_attrs",
    "has_set_attrs",
    "has_mutable_set_attrs",
    "has_mapping_attrs",
    "has_mutable_mapping_attrs",
    "has_mapping_view_attrs",
    "has_item_view_attrs",
    "has_keys_view_attrs",
    "has_values_view_attrs",
    "has_awaitable_attrs",
    "has_coroutine_attrs",
    "has_async_iterable_attrs",
    "has_async_iterator_attrs",
    "has_async_generator_attrs",
]



"""Type definitions and mappings for collections.abc types.

This module provides enums and mappings between abstract base classes
from collections.abc and typing modules to their corresponding attribute sets.
"""

import typing
from collections import abc
from enum import Enum
from typing import Any, Dict
from typing import Set as TypingSet

from hasattrs.attributes import (
    ASYNC_GENERATOR,
    ASYNC_ITERABLE,
    ASYNC_ITERATOR,
    AWAITABLE,
    BYTE_STRING,
    CALLABLE,
    COLLECTION,
    CONTAINER,
    COROUTINE,
    GENERATOR,
    HASHABLE,
    ITEMS_VIEW,
    ITERABLE,
    ITERATOR,
    KEYS_VIEW,
    MAPPING,
    MAPPING_VIEW,
    MUTABLE_MAPPING,
    MUTABLE_SEQUENCE,
    MUTABLE_SET,
    REVERSIBLE,
    SEQUENCE,
    SET,
    SIZED,
    VALUES_VIEW,
)


class ABC(Enum):
    """Enum of collections.abc abstract base classes."""

    Container = abc.Container
    Hashable = abc.Hashable
    Iterable = abc.Iterable
    Iterator = abc.Iterator
    Reversible = abc.Reversible
    Generator = abc.Generator
    Sized = abc.Sized
    Callable = abc.Callable
    Collection = abc.Collection
    Sequence = abc.Sequence
    MutableSequence = abc.MutableSequence
    ByteString = abc.ByteString
    Set = abc.Set
    MutableSet = abc.MutableSet
    Mapping = abc.Mapping
    MutableMapping = abc.MutableMapping
    MappingView = abc.MappingView
    ItemsView = abc.ItemsView
    KeysView = abc.KeysView
    ValuesView = abc.ValuesView
    Awaitable = abc.Awaitable
    Coroutine = abc.Coroutine
    AsyncIterable = abc.AsyncIterable
    AsyncIterator = abc.AsyncIterator
    AsyncGenerator = abc.AsyncGenerator


class Typing(Enum):
    """Enum of typing module abstract base classes."""

    Container = typing.Container
    Hashable = typing.Hashable
    Iterable = typing.Iterable
    Iterator = typing.Iterator
    Reversible = typing.Reversible
    Generator = typing.Generator
    Sized = typing.Sized
    Callable = typing.Callable
    Collection = typing.Collection
    Sequence = typing.Sequence
    MutableSequence = typing.MutableSequence
    ByteString = typing.ByteString
    Set = typing.Set
    MutableSet = typing.MutableSet
    Mapping = typing.Mapping
    MutableMapping = typing.MutableMapping
    MappingView = typing.MappingView
    ItemsView = typing.ItemsView
    KeysView = typing.KeysView
    ValuesView = typing.ValuesView
    Awaitable = typing.Awaitable
    Coroutine = typing.Coroutine
    AsyncIterable = typing.AsyncIterable
    AsyncIterator = typing.AsyncIterator
    AsyncGenerator = typing.AsyncGenerator


abc_attrs: Dict[Any, TypingSet[str]] = {
    abc.Container: CONTAINER,
    typing.Container: CONTAINER,
    "Container": CONTAINER,
    abc.Hashable: HASHABLE,
    typing.Hashable: HASHABLE,
    "Hashable": HASHABLE,
    abc.Iterable: ITERABLE,
    typing.Iterable: ITERABLE,
    "Iterable": ITERABLE,
    abc.Iterator: ITERATOR,
    typing.Iterator: ITERATOR,
    "Iterator": ITERATOR,
    abc.Reversible: REVERSIBLE,
    typing.Reversible: REVERSIBLE,
    "Reversible": REVERSIBLE,
    abc.Generator: GENERATOR,
    typing.Generator: GENERATOR,
    "Generator": GENERATOR,
    abc.Sized: SIZED,
    typing.Sized: SIZED,
    "Sized": SIZED,
    abc.Callable: CALLABLE,
    typing.Callable: CALLABLE,
    "Callable": CALLABLE,
    abc.Collection: COLLECTION,
    typing.Collection: COLLECTION,
    "Collection": COLLECTION,
    abc.Sequence: SEQUENCE,
    typing.Sequence: SEQUENCE,
    "Sequence": SEQUENCE,
    abc.MutableSequence: MUTABLE_SEQUENCE,
    typing.MutableSequence: MUTABLE_SEQUENCE,
    "MutableSequence": MUTABLE_SEQUENCE,
    abc.ByteString: BYTE_STRING,
    typing.ByteString: BYTE_STRING,
    "ByteString": BYTE_STRING,
    abc.Set: SET,
    typing.Set: SET,
    "Set": SET,
    abc.MutableSet: MUTABLE_SET,
    typing.MutableSet: MUTABLE_SET,
    "MutableSet": MUTABLE_SET,
    abc.Mapping: MAPPING,
    typing.Mapping: MAPPING,
    "Mapping": MAPPING,
    abc.MutableMapping: MUTABLE_MAPPING,
    typing.MutableMapping: MUTABLE_MAPPING,
    "MutableMapping": MUTABLE_MAPPING,
    abc.MappingView: MAPPING_VIEW,
    typing.MappingView: MAPPING_VIEW,
    "MappingView": MAPPING_VIEW,
    abc.ItemsView: ITEMS_VIEW,
    typing.ItemsView: ITEMS_VIEW,
    "ItemsView": ITEMS_VIEW,
    abc.KeysView: KEYS_VIEW,
    typing.KeysView: KEYS_VIEW,
    "KeysView": KEYS_VIEW,
    abc.ValuesView: VALUES_VIEW,
    typing.ValuesView: VALUES_VIEW,
    "ValuesView": VALUES_VIEW,
    abc.Awaitable: AWAITABLE,
    typing.Awaitable: AWAITABLE,
    "Awaitable": AWAITABLE,
    abc.Coroutine: COROUTINE,
    typing.Coroutine: COROUTINE,
    "Coroutine": COROUTINE,
    abc.AsyncIterable: ASYNC_ITERABLE,
    typing.AsyncIterable: ASYNC_ITERABLE,
    "AsyncIterable": ASYNC_ITERABLE,
    abc.AsyncIterator: ASYNC_ITERATOR,
    typing.AsyncIterator: ASYNC_ITERATOR,
    "AsyncIterator": ASYNC_ITERATOR,
    abc.AsyncGenerator: ASYNC_GENERATOR,
    typing.AsyncGenerator: ASYNC_GENERATOR,
    "AsyncGenerator": ASYNC_GENERATOR,
}



"""Tests for attribute set definitions."""

from hasattrs.attributes import (
    ASYNC_GENERATOR,
    ASYNC_ITERABLE,
    ASYNC_ITERATOR,
    AWAITABLE,
    CALLABLE,
    COLLECTION,
    CONTAINER,
    COROUTINE,
    GENERATOR,
    HASHABLE,
    ITEMS_VIEW,
    ITERABLE,
    ITERATOR,
    KEYS_VIEW,
    MAPPING,
    MAPPING_VIEW,
    MUTABLE_MAPPING,
    MUTABLE_SEQUENCE,
    MUTABLE_SET,
    REVERSIBLE,
    SEQUENCE,
    SET,
    SIZED,
    VALUES_VIEW,
)


class TestAttributeSets:
    """Test that attribute sets contain the expected attributes."""

    def test_container_attrs(self):
        assert "__contains__" in CONTAINER

    def test_hashable_attrs(self):
        assert "__hash__" in HASHABLE

    def test_iterable_attrs(self):
        assert "__iter__" in ITERABLE

    def test_iterator_attrs(self):
        assert "__iter__" in ITERATOR
        assert "__next__" in ITERATOR

    def test_reversible_attrs(self):
        assert "__iter__" in REVERSIBLE
        assert "__reversed__" in REVERSIBLE

    def test_generator_attrs(self):
        assert "__iter__" in GENERATOR
        assert "__next__" in GENERATOR
        assert "send" in GENERATOR
        assert "throw" in GENERATOR
        assert "close" in GENERATOR

    def test_sized_attrs(self):
        assert "__len__" in SIZED

    def test_callable_attrs(self):
        assert "__call__" in CALLABLE

    def test_collection_attrs(self):
        assert "__len__" in COLLECTION
        assert "__iter__" in COLLECTION
        assert "__contains__" in COLLECTION

    def test_sequence_attrs(self):
        """Test that Sequence has all required attributes including __getitem__."""
        assert "__getitem__" in SEQUENCE
        assert "__len__" in SEQUENCE
        assert "__iter__" in SEQUENCE
        assert "__contains__" in SEQUENCE
        assert "__reversed__" in SEQUENCE
        assert "index" in SEQUENCE
        assert "count" in SEQUENCE

    def test_mutable_sequence_attrs(self):
        assert "__setitem__" in MUTABLE_SEQUENCE
        assert "__delitem__" in MUTABLE_SEQUENCE
        assert "insert" in MUTABLE_SEQUENCE
        assert "append" in MUTABLE_SEQUENCE
        assert "reverse" in MUTABLE_SEQUENCE
        assert "extend" in MUTABLE_SEQUENCE
        assert "pop" in MUTABLE_SEQUENCE
        assert "remove" in MUTABLE_SEQUENCE
        assert "__iadd__" in MUTABLE_SEQUENCE

    def test_set_attrs(self):
        assert "__le__" in SET
        assert "__lt__" in SET
        assert "__eq__" in SET
        assert "__ne__" in SET
        assert "__gt__" in SET
        assert "__ge__" in SET
        assert "isdisjoint" in SET

    def test_mutable_set_attrs(self):
        assert "add" in MUTABLE_SET
        assert "discard" in MUTABLE_SET
        assert "clear" in MUTABLE_SET
        assert "pop" in MUTABLE_SET
        assert "remove" in MUTABLE_SET

    def test_mapping_attrs(self):
        """Test that Mapping has all required attributes including __getitem__."""
        assert "__getitem__" in MAPPING
        assert "__len__" in MAPPING
        assert "__iter__" in MAPPING
        assert "__contains__" in MAPPING
        assert "keys" in MAPPING
        assert "items" in MAPPING
        assert "values" in MAPPING
        assert "get" in MAPPING
        assert "__eq__" in MAPPING
        assert "__ne__" in MAPPING

    def test_mutable_mapping_attrs(self):
        assert "__setitem__" in MUTABLE_MAPPING
        assert "__delitem__" in MUTABLE_MAPPING
        assert "pop" in MUTABLE_MAPPING
        assert "popitem" in MUTABLE_MAPPING
        assert "clear" in MUTABLE_MAPPING
        assert "update" in MUTABLE_MAPPING
        assert "setdefault" in MUTABLE_MAPPING

    def test_mapping_view_attrs(self):
        assert "__len__" in MAPPING_VIEW

    def test_items_view_attrs(self):
        assert "__len__" in ITEMS_VIEW

    def test_keys_view_attrs(self):
        assert "__len__" in KEYS_VIEW

    def test_values_view_attrs(self):
        assert "__len__" in VALUES_VIEW

    def test_awaitable_attrs(self):
        assert "__await__" in AWAITABLE

    def test_coroutine_attrs(self):
        assert "__await__" in COROUTINE
        assert "send" in COROUTINE
        assert "throw" in COROUTINE
        assert "close" in COROUTINE

    def test_async_iterable_attrs(self):
        assert "__aiter__" in ASYNC_ITERABLE

    def test_async_iterator_attrs(self):
        assert "__aiter__" in ASYNC_ITERATOR
        assert "__anext__" in ASYNC_ITERATOR

    def test_async_generator_attrs(self):
        assert "__aiter__" in ASYNC_GENERATOR
        assert "__anext__" in ASYNC_GENERATOR
        assert "asend" in ASYNC_GENERATOR
        assert "athrow" in ASYNC_GENERATOR
        assert "aclose" in ASYNC_GENERATOR



"""Attribute set definitions for collections.abc types.

This module defines sets of attribute names (dunder methods and regular methods)
that correspond to the abstract base classes in collections.abc and typing.
"""

CONTAINER = {"__contains__"}
HASHABLE = {"__hash__"}
ITERABLE = {"__iter__"}
ITERATOR = ITERABLE | {"__next__"}
REVERSIBLE = ITERABLE | {"__reversed__"}
GENERATOR = ITERATOR | {"send", "throw", "close"}
SIZED = {"__len__"}
CALLABLE = {"__call__"}
COLLECTION = SIZED | ITERABLE | CONTAINER
SEQUENCE = REVERSIBLE | COLLECTION | {"__getitem__", "index", "count"}
MUTABLE_SEQUENCE = SEQUENCE | {
    "__setitem__",
    "__delitem__",
    "insert",
    "append",
    "reverse",
    "extend",
    "pop",
    "remove",
    "__iadd__",
}
BYTE_STRING = SEQUENCE
SET = COLLECTION | {
    "__le__",
    "__lt__",
    "__eq__",
    "__ne__",
    "__gt__",
    "__ge__",
    "__and__",
    "__or__",
    "__sub__",
    "__xor__",
    "isdisjoint",
}
MUTABLE_SET = SET | {
    "add",
    "discard",
    "clear",
    "pop",
    "remove",
    "__ior__",
    "__iand__",
    "__ixor__",
    "__isub__",
}
MAPPING = COLLECTION | {"__getitem__", "keys", "items", "values", "get", "__eq__", "__ne__"}
MUTABLE_MAPPING = MAPPING | {
    "__setitem__",
    "__delitem__",
    "pop",
    "popitem",
    "clear",
    "update",
    "setdefault",
}
MAPPING_VIEW = SIZED
ITEMS_VIEW = MAPPING_VIEW | SET
KEYS_VIEW = MAPPING_VIEW | COLLECTION
VALUES_VIEW = MAPPING_VIEW | COLLECTION
AWAITABLE = {"__await__"}
COROUTINE = AWAITABLE | {"send", "throw", "close"}
ASYNC_ITERABLE = {"__aiter__"}
ASYNC_ITERATOR = ASYNC_ITERABLE | {"__anext__"}
ASYNC_GENERATOR = ASYNC_ITERATOR | {"asend", "athrow", "aclose"}



"""Tests for attribute checker functions."""

import contextlib
import warnings
from collections import abc
from typing import Mapping as TypingMapping

from hasattrs import (
    has_abc_attrs,
    has_async_generator_attrs,
    has_async_iterable_attrs,
    has_async_iterator_attrs,
    has_attrs,
    has_awaitable_attrs,
    has_byte_string_attrs,
    has_callable_attrs,
    has_collection_attrs,
    has_container_attrs,
    has_coroutine_attrs,
    has_generator_attrs,
    has_hashable_attrs,
    has_item_view_attrs,
    has_iterable_attrs,
    has_iterator_attrs,
    has_keys_view_attrs,
    has_mapping_attrs,
    has_mapping_view_attrs,
    has_mutable_mapping_attrs,
    has_mutable_sequence_attrs,
    has_mutable_set_attrs,
    has_reversible_attrs,
    has_sequence_attrs,
    has_set_attrs,
    has_sized_attrs,
    has_values_view_attrs,
)


class TestHasAttrs:
    """Test the basic has_attrs function."""

    def test_has_attrs_true(self):
        assert has_attrs([], ["__len__", "__iter__"])

    def test_has_attrs_false(self):
        assert not has_attrs(5, ["__len__"])

    def test_has_attrs_empty(self):
        assert has_attrs(None, [])

    def test_has_attrs_partial(self):
        assert not has_attrs([1, 2], ["__len__", "__missing_attr__"])


class TestHasAbcAttrs:
    """Test the generic has_abc_attrs function."""

    def test_with_abc_class(self):
        assert has_abc_attrs({}, abc.Mapping)

    def test_with_typing_class(self):
        assert has_abc_attrs({}, TypingMapping)

    def test_with_string(self):
        assert has_abc_attrs({}, "Mapping")

    def test_fails_correctly(self):
        assert not has_abc_attrs([], abc.Mapping)


class TestContainerAttrs:
    """Test has_container_attrs function."""

    def test_list(self):
        assert has_container_attrs([1, 2, 3])

    def test_dict(self):
        assert has_container_attrs({})

    def test_set(self):
        assert has_container_attrs(set())

    def test_int(self):
        assert not has_container_attrs(5)


class TestHashableAttrs:
    """Test has_hashable_attrs function."""

    def test_int(self):
        assert has_hashable_attrs(5)

    def test_str(self):
        assert has_hashable_attrs("hello")

    def test_tuple(self):
        assert has_hashable_attrs((1, 2, 3))

    def test_list(self):
        # Lists have __hash__ attribute (set to None), so they pass the attribute check
        # This is a quirk of Python - the attribute exists but is set to None
        assert has_hashable_attrs([1, 2, 3])


class TestIterableAttrs:
    """Test has_iterable_attrs function."""

    def test_list(self):
        assert has_iterable_attrs([1, 2, 3])

    def test_dict(self):
        assert has_iterable_attrs({})

    def test_str(self):
        assert has_iterable_attrs("hello")

    def test_int(self):
        assert not has_iterable_attrs(5)


class TestIteratorAttrs:
    """Test has_iterator_attrs function."""

    def test_list_iterator(self):
        assert has_iterator_attrs(iter([1, 2, 3]))

    def test_list(self):
        assert not has_iterator_attrs([1, 2, 3])


class TestReversibleAttrs:
    """Test has_reversible_attrs function."""

    def test_list(self):
        assert has_reversible_attrs([1, 2, 3])

    def test_tuple(self):
        # In Python 3.8, tuples don't have __reversed__ method
        # They use the default reversed() which doesn't require the method
        assert not has_reversible_attrs((1, 2, 3))

    def test_dict(self):
        assert has_reversible_attrs({})

    def test_set(self):
        # Sets are not reversible
        assert not has_reversible_attrs(set())


class TestGeneratorAttrs:
    """Test has_generator_attrs function."""

    def test_generator(self):
        def gen():
            yield 1

        assert has_generator_attrs(gen())

    def test_list(self):
        assert not has_generator_attrs([1, 2, 3])


class TestSizedAttrs:
    """Test has_sized_attrs function."""

    def test_list(self):
        assert has_sized_attrs([1, 2, 3])

    def test_dict(self):
        assert has_sized_attrs({})

    def test_str(self):
        assert has_sized_attrs("hello")

    def test_int(self):
        assert not has_sized_attrs(5)


class TestCallableAttrs:
    """Test has_callable_attrs function."""

    def test_function(self):
        def f():
            pass

        assert has_callable_attrs(f)

    def test_lambda(self):
        assert has_callable_attrs(lambda x: x)

    def test_callable_class(self):
        class C:
            def __call__(self):
                pass

        assert has_callable_attrs(C())

    def test_int(self):
        assert not has_callable_attrs(5)


class TestCollectionAttrs:
    """Test has_collection_attrs function."""

    def test_list(self):
        assert has_collection_attrs([1, 2, 3])

    def test_dict(self):
        assert has_collection_attrs({})

    def test_set(self):
        assert has_collection_attrs(set())

    def test_str(self):
        assert has_collection_attrs("hello")

    def test_int(self):
        assert not has_collection_attrs(5)


class TestSequenceAttrs:
    """Test has_sequence_attrs function."""

    def test_list(self):
        assert has_sequence_attrs([1, 2, 3])

    def test_tuple(self):
        # Tuples don't have __reversed__ in Python 3.8
        assert not has_sequence_attrs((1, 2, 3))

    def test_str(self):
        # Strings don't have __reversed__ in Python 3.8
        assert not has_sequence_attrs("hello")

    def test_dict(self):
        assert not has_sequence_attrs({})

    def test_set(self):
        assert not has_sequence_attrs(set())


class TestMutableSequenceAttrs:
    """Test has_mutable_sequence_attrs function."""

    def test_list(self):
        assert has_mutable_sequence_attrs([1, 2, 3])

    def test_tuple(self):
        assert not has_mutable_sequence_attrs((1, 2, 3))


class TestByteStringAttrs:
    """Test has_byte_string_attrs function (deprecated)."""

    def test_bytes(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # bytes don't have __reversed__ in Python 3.8, so they don't pass Sequence check
            result = has_byte_string_attrs(b"hello")
            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)
            assert "deprecated" in str(w[0].message).lower()
            assert result is False

    def test_bytearray(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            has_byte_string_attrs(bytearray(b"hello"))
            assert len(w) == 1
            assert issubclass(w[0].category, DeprecationWarning)


class TestSetAttrs:
    """Test has_set_attrs function."""

    def test_set(self):
        assert has_set_attrs({1, 2, 3})

    def test_frozenset(self):
        assert has_set_attrs(frozenset([1, 2, 3]))

    def test_list(self):
        assert not has_set_attrs([1, 2, 3])


class TestMutableSetAttrs:
    """Test has_mutable_set_attrs function."""

    def test_set(self):
        assert has_mutable_set_attrs({1, 2, 3})

    def test_frozenset(self):
        assert not has_mutable_set_attrs(frozenset([1, 2, 3]))


class TestMappingAttrs:
    """Test has_mapping_attrs function."""

    def test_dict(self):
        assert has_mapping_attrs({})

    def test_list(self):
        assert not has_mapping_attrs([])

    def test_custom_mapping(self):
        class MyMapping:
            def __getitem__(self, key):
                pass

            def __iter__(self):
                pass

            def __len__(self):
                pass

            def __contains__(self, key):
                pass

            def keys(self):
                pass

            def items(self):
                pass

            def values(self):
                pass

            def get(self, key):
                pass

            def __eq__(self, other):
                pass

            def __ne__(self, other):
                pass

        assert has_mapping_attrs(MyMapping())


class TestMutableMappingAttrs:
    """Test has_mutable_mapping_attrs function."""

    def test_dict(self):
        assert has_mutable_mapping_attrs({})

    def test_tuple(self):
        assert not has_mutable_mapping_attrs((1, 2))


class TestMappingViewAttrs:
    """Test has_mapping_view_attrs function."""

    def test_dict_keys(self):
        assert has_mapping_view_attrs({}.keys())

    def test_dict_values(self):
        assert has_mapping_view_attrs({}.values())

    def test_dict_items(self):
        assert has_mapping_view_attrs({}.items())


class TestItemViewAttrs:
    """Test has_item_view_attrs function."""

    def test_dict_items(self):
        assert has_item_view_attrs({}.items())

    def test_list(self):
        assert not has_item_view_attrs([])


class TestKeysViewAttrs:
    """Test has_keys_view_attrs function."""

    def test_dict_keys(self):
        assert has_keys_view_attrs({}.keys())

    def test_list(self):
        # Lists have __len__, __iter__, __contains__ which satisfies KEYS_VIEW (MAPPING_VIEW | COLLECTION)
        assert has_keys_view_attrs([])


class TestValuesViewAttrs:
    """Test has_values_view_attrs function."""

    def test_dict_values(self):
        # dict_values don't have __contains__ in any Python version (as of 3.12)
        # ValuesView requires __len__, __iter__, and __contains__,
        # but dict_values only has __len__ and __iter__
        assert not has_values_view_attrs({}.values())

    def test_list(self):
        # Lists have __len__, __iter__, __contains__ which satisfies VALUES_VIEW (MAPPING_VIEW | COLLECTION)
        assert has_values_view_attrs([])


class TestAwaitableAttrs:
    """Test has_awaitable_attrs function."""

    def test_coroutine(self):
        async def f():
            pass

        coro = f()
        try:
            assert has_awaitable_attrs(coro)
        finally:
            coro.close()

    def test_int(self):
        assert not has_awaitable_attrs(5)


class TestCoroutineAttrs:
    """Test has_coroutine_attrs function."""

    def test_coroutine(self):
        async def f():
            pass

        coro = f()
        try:
            assert has_coroutine_attrs(coro)
        finally:
            coro.close()

    def test_int(self):
        assert not has_coroutine_attrs(5)


class TestAsyncIterableAttrs:
    """Test has_async_iterable_attrs function."""

    def test_async_generator(self):
        async def gen():
            yield 1

        assert has_async_iterable_attrs(gen())

    def test_list(self):
        assert not has_async_iterable_attrs([])


class TestAsyncIteratorAttrs:
    """Test has_async_iterator_attrs function."""

    def test_async_generator(self):
        async def gen():
            yield 1

        assert has_async_iterator_attrs(gen())

    def test_list(self):
        assert not has_async_iterator_attrs([])


class TestAsyncGeneratorAttrs:
    """Test has_async_generator_attrs function."""

    def test_async_generator(self):
        async def gen():
            yield 1

        agen = gen()
        try:
            assert has_async_generator_attrs(agen)
        finally:
            # Clean up async generator
            with contextlib.suppress(Exception):
                agen.aclose()

    def test_list(self):
        assert not has_async_generator_attrs([])


class TestCustomClasses:
    """Test with custom classes that implement protocols."""

    def test_custom_sequence(self):
        class MySequence:
            def __getitem__(self, index):
                pass

            def __len__(self):
                pass

            def __contains__(self, value):
                pass

            def __iter__(self):
                pass

            def __reversed__(self):
                pass

            def index(self, value):
                pass

            def count(self, value):
                pass

        assert has_sequence_attrs(MySequence())

    def test_custom_mapping(self):
        class MyMapping:
            def __getitem__(self, key):
                pass

            def __iter__(self):
                pass

            def __len__(self):
                pass

            def __contains__(self, key):
                pass

            def keys(self):
                pass

            def items(self):
                pass

            def values(self):
                pass

            def get(self, key, default=None):
                pass

            def __eq__(self, other):
                pass

            def __ne__(self, other):
                pass

        assert has_mapping_attrs(MyMapping())
        assert not has_mutable_mapping_attrs(MyMapping())

    def test_incomplete_protocol(self):
        class IncompleteMapping:
            def __getitem__(self, key):
                pass

            def __len__(self):
                pass

            # Missing __iter__, keys, items, values, get, etc.

        assert not has_mapping_attrs(IncompleteMapping())



"""Attribute checker functions for collections.abc types.

This module provides functions to check if objects have the required attributes
for various abstract base classes from collections.abc and typing.

Example:
    >>> from hasattrs import has_mapping_attrs
    >>> class MyDict:
    ...     def __getitem__(self, key): pass
    ...     def __iter__(self): pass
    ...     def __len__(self): pass
    ...     # ... other Mapping methods
    >>> has_mapping_attrs(MyDict())
    True
"""

import warnings
from typing import Any, Iterable

from hasattrs.attributes import (
    ASYNC_GENERATOR,
    ASYNC_ITERABLE,
    ASYNC_ITERATOR,
    AWAITABLE,
    BYTE_STRING,
    CALLABLE,
    COLLECTION,
    CONTAINER,
    COROUTINE,
    GENERATOR,
    HASHABLE,
    ITEMS_VIEW,
    ITERABLE,
    ITERATOR,
    KEYS_VIEW,
    MAPPING,
    MAPPING_VIEW,
    MUTABLE_MAPPING,
    MUTABLE_SEQUENCE,
    MUTABLE_SET,
    REVERSIBLE,
    SEQUENCE,
    SET,
    SIZED,
    VALUES_VIEW,
)
from hasattrs.types import abc_attrs


def has_attrs(obj: Any, attrs: Iterable[str]) -> bool:
    """Check if an object has all specified attributes.

    Args:
        obj: The object to check.
        attrs: An iterable of attribute names (strings) to check for.

    Returns:
        True if the object has all specified attributes, False otherwise.

    Example:
        >>> has_attrs([], ['__len__', '__iter__'])
        True
        >>> has_attrs(5, ['__len__'])
        False
    """
    return all(hasattr(obj, a) for a in attrs)


def has_abc_attrs(obj: Any, abc: Any) -> bool:
    """Check if an object has all attributes for a collections.abc type.

    Args:
        obj: The object to check.
        abc: The abstract base class to check against. Can be:
            - A collections.abc class (e.g., abc.Mapping)
            - A typing class (e.g., typing.Mapping)
            - A string name (e.g., 'Mapping')

    Returns:
        True if the object has all required attributes, False otherwise.

    Example:
        >>> from collections.abc import Mapping
        >>> has_abc_attrs({}, Mapping)
        True
        >>> has_abc_attrs([], Mapping)
        False
    """
    return has_attrs(obj, abc_attrs[abc])


def has_container_attrs(obj: Any) -> bool:
    """Check if an object has Container attributes (__contains__)."""
    return has_attrs(obj, CONTAINER)


def has_hashable_attrs(obj: Any) -> bool:
    """Check if an object has Hashable attributes (__hash__)."""
    return has_attrs(obj, HASHABLE)


def has_iterable_attrs(obj: Any) -> bool:
    """Check if an object has Iterable attributes (__iter__)."""
    return has_attrs(obj, ITERABLE)


def has_iterator_attrs(obj: Any) -> bool:
    """Check if an object has Iterator attributes (__iter__, __next__)."""
    return has_attrs(obj, ITERATOR)


def has_reversible_attrs(obj: Any) -> bool:
    """Check if an object has Reversible attributes (__iter__, __reversed__)."""
    return has_attrs(obj, REVERSIBLE)


def has_generator_attrs(obj: Any) -> bool:
    """Check if an object has Generator attributes."""
    return has_attrs(obj, GENERATOR)


def has_sized_attrs(obj: Any) -> bool:
    """Check if an object has Sized attributes (__len__)."""
    return has_attrs(obj, SIZED)


def has_callable_attrs(obj: Any) -> bool:
    """Check if an object has Callable attributes (__call__)."""
    return has_attrs(obj, CALLABLE)


def has_collection_attrs(obj: Any) -> bool:
    """Check if an object has Collection attributes."""
    return has_attrs(obj, COLLECTION)


def has_sequence_attrs(obj: Any) -> bool:
    """Check if an object has Sequence attributes."""
    return has_attrs(obj, SEQUENCE)


def has_mutable_sequence_attrs(obj: Any) -> bool:
    """Check if an object has MutableSequence attributes."""
    return has_attrs(obj, MUTABLE_SEQUENCE)


def has_byte_string_attrs(obj: Any) -> bool:
    """Check if an object has ByteString attributes.

    .. deprecated:: 0.1.0
        ByteString is deprecated in Python 3.9+ and will be removed in Python 3.14.
        Use collections.abc.Sequence with isinstance checks instead.
    """
    warnings.warn(
        "has_byte_string_attrs is deprecated because collections.abc.ByteString "
        "is deprecated since Python 3.9 and will be removed in Python 3.14. "
        "Use has_sequence_attrs or isinstance checks instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return has_attrs(obj, BYTE_STRING)


def has_set_attrs(obj: Any) -> bool:
    """Check if an object has Set attributes."""
    return has_attrs(obj, SET)


def has_mutable_set_attrs(obj: Any) -> bool:
    """Check if an object has MutableSet attributes."""
    return has_attrs(obj, MUTABLE_SET)


def has_mapping_attrs(obj: Any) -> bool:
    """Check if an object has Mapping attributes."""
    return has_attrs(obj, MAPPING)


def has_mutable_mapping_attrs(obj: Any) -> bool:
    """Check if an object has MutableMapping attributes."""
    return has_attrs(obj, MUTABLE_MAPPING)


def has_mapping_view_attrs(obj: Any) -> bool:
    """Check if an object has MappingView attributes."""
    return has_attrs(obj, MAPPING_VIEW)


def has_item_view_attrs(obj: Any) -> bool:
    """Check if an object has ItemsView attributes."""
    return has_attrs(obj, ITEMS_VIEW)


def has_keys_view_attrs(obj: Any) -> bool:
    """Check if an object has KeysView attributes."""
    return has_attrs(obj, KEYS_VIEW)


def has_values_view_attrs(obj: Any) -> bool:
    """Check if an object has ValuesView attributes."""
    return has_attrs(obj, VALUES_VIEW)


def has_awaitable_attrs(obj: Any) -> bool:
    """Check if an object has Awaitable attributes (__await__)."""
    return has_attrs(obj, AWAITABLE)


def has_coroutine_attrs(obj: Any) -> bool:
    """Check if an object has Coroutine attributes."""
    return has_attrs(obj, COROUTINE)


def has_async_iterable_attrs(obj: Any) -> bool:
    """Check if an object has AsyncIterable attributes (__aiter__)."""
    return has_attrs(obj, ASYNC_ITERABLE)


def has_async_iterator_attrs(obj: Any) -> bool:
    """Check if an object has AsyncIterator attributes."""
    return has_attrs(obj, ASYNC_ITERATOR)


def has_async_generator_attrs(obj: Any) -> bool:
    """Check if an object has AsyncGenerator attributes."""
    return has_attrs(obj, ASYNC_GENERATOR)
